#ifndef __IO_
#define __IO_

#include <windows.h>
#include <string.h>
#include <math.h>

class OutStream
{
	private:
		long long	 currentfilepos;
		int		 stream;
		long long	 maxsize;
		unsigned char	*data;
		bool		 pbd;
		bool		 holdpbd;
		int		 pbdlen;
		int		 pbdbuffer[128];
		inline int	 GetByte	(long long, int);
		inline bool	 GetBit		(long long, int);
		void		 InitPBD	();
		void		 CompletePBD	();
	public:
		void		 OutputNumber		(long long, int);
		void		 OutputNumberRaw	(long long, int);
		void		 OutputNumberPBD	(long long, int);
		void		 OutputString		(char *);
		long long	 SetPos			(long long);
		long long	 ActPos			();
		long long	 MaxSize		();
		OutStream				(char *, long long);
		~OutStream				();
};

OutStream::OutStream(char *filename, long long max)
{
	maxsize		= max;
	currentfilepos	= 0;
	stream		= _lcreat(filename, 0);
	data			= new unsigned char [maxsize];
	pbd		= 0;
	holdpbd		= 0;
}

OutStream::~OutStream()
{
	if (pbd) CompletePBD();

	_hwrite(stream, data, currentfilepos);
	_lclose(stream);
	delete [] data;
	data = 0;
}

inline int OutStream::GetByte(long long number, int byte)
{
	if (byte > 7) return -1;

	return (number >> (8 * byte)) % 256;
}

inline bool OutStream::GetBit(long long number, int bit)
{
	if (bit > 63) return -1;

	return (number >> bit) % 2;
}

void OutStream::OutputNumber(long long number, int bytes)
{
	if (pbd && !holdpbd) CompletePBD();

	if (bytes > 8) return;

	for (int i = 0; i < bytes; i++)
	{
		data[currentfilepos] = GetByte(number, i);
		currentfilepos++;
	}
}

void OutStream::OutputNumberRaw(long long number, int bytes)
{
	if (pbd && !holdpbd) CompletePBD();

	if (bytes > 8) return;

	for (int i = bytes-1; i >= 0; i--)
	{
		data[currentfilepos] = GetByte(number, i);
		currentfilepos++;
	}
}

void OutStream::OutputNumberPBD(long long number, int bits)
{
	if (!pbd) InitPBD();

	int	out;

	for (int i = 0; i < bits; i++)
	{
		pbdbuffer[pbdlen] = GetBit(number, i);
		pbdlen++;
	}
	while (pbdlen >= 8)
	{
		out = 0;
		for (int i = 0; i < 8; i++)
		{
			out = out + (pbdbuffer[i]*pow(2, i));
		}
		pbdlen = pbdlen - 8;
		for (int i = 0; i < pbdlen; i++)
		{
			pbdbuffer[i] = pbdbuffer[i+8];
		}
		holdpbd = true;
		OutputNumber(out, 1);
		holdpbd = false;
	}
}

void OutStream::OutputString(char *string)
{
	if (pbd && !holdpbd) CompletePBD();

	for (int i = 0; i < strlen(string); i++)
	{
		OutputNumber(string[i], 1);
	}
}

void OutStream::InitPBD()
{
	pbdlen = 0;
	pbd = 1;
}

void OutStream::CompletePBD()
{
	if (pbdlen > 0)
	{
		int	out = 0;
		for (int i = 0; i < 8; i++)
		{
			if (i < pbdlen) out = out + (pbdbuffer[i]*pow(2, i));
		}
		holdpbd = true;
		OutputNumber(out, 1);
		holdpbd = false;
	}

	pbd = 0;
}

long long OutStream::SetPos(long long pos)
{
	if (pbd && !holdpbd) CompletePBD();

	long long	oldpos = currentfilepos;

	currentfilepos = pos;

	return oldpos;
}

long long OutStream::ActPos()
{
	return currentfilepos;
}

long long OutStream::MaxSize()
{
	return maxsize;
}

class InStream
{
	private:
		long long	 currentfilepos;
		int		 stream;
		long long	 size;
		unsigned char	*data;
		bool		 pbd;
		bool		 holdpbd;
		int		 pbdlen;
		bool		 pbdbuffer[128];
		inline bool	 GetBit		(long long, int);
		void		 InitPBD	();
		void		 CompletePBD	();
	public:
		long long	 InputNumber	(int);
		long long	 InputNumberRaw	(int);
		long long	 InputNumberPBD	(int);
		char		*InputString	(int);
		long long	 SetPos		(long long);
		long long	 ActPos		();
		long long	 Size		();
		InStream			(char *);
		~InStream			();
};

InStream::InStream(char *filename)
{
	int	 gfsizebuf;

	currentfilepos	= 0;
	stream		= _lopen(filename, OF_READ);
	size			= GetFileSize(stream, &gfsizebuf);
	data			= new unsigned char [size];
	pbd		= 0;
	holdpbd		= 0;
	_hread(stream, data, size);
	_lclose(stream);
}

InStream::~InStream()
{
	if (pbd) CompletePBD();

	delete [] data;
	data = 0;
}

inline bool InStream::GetBit(long long number, int bit)
{
	if (bit > 63) return -1;

	return (number >> bit) % 2;
}

long long InStream::InputNumber(int bytes)
{
	if (pbd && !holdpbd) CompletePBD();

	if (bytes > 8) return -1;

	long long	 rval = 0;

	for (int i = 0; i < bytes; i++)
	{
		rval = rval + data[currentfilepos]*pow(2, i * 8);
		currentfilepos++;
	}

	return rval;
}

long long InStream::InputNumberRaw(int bytes)
{
	if (pbd && !holdpbd) CompletePBD();

	if (bytes > 8) return -1;

	long long	 rval = 0;

	for (int i = bytes-1; i >= 0; i--)
	{
		rval = rval + data[currentfilepos]*pow(2, i * 8);
		currentfilepos++;
	}

	return rval;
}

long long InStream::InputNumberPBD(int bits)
{
	if (!pbd) InitPBD();

	int	inp;
	int	ret=0;

	while (pbdlen < bits)
	{
		holdpbd = true;
		inp = InputNumber(1);
		holdpbd = false;
		for (int i = 0; i < 8; i++)
		{
			pbdbuffer[pbdlen] = GetBit(inp, i);
			pbdlen++;
		}
	}
	for (int i = 0; i < bits; i++)
	{
		ret = ret + (pbdbuffer[i]*pow(2, i));
	}
	pbdlen = pbdlen-bits;
	for (int i = 0; i < pbdlen; i++)
	{
		pbdbuffer[i] = pbdbuffer[i+bits];
	}

	return ret;
}

char *InStream::InputString(int bytes)
{
	if (pbd && !holdpbd) CompletePBD();

	char	 *rval = new char [bytes+1];

	for (int i = 0; i < bytes; i++)
	{
		rval[i] = InputNumber(1);
	}
	rval[bytes] = 0;

	return rval;
}

void InStream::InitPBD()
{
	pbdlen = 0;
	pbd = 1;
}

void InStream::CompletePBD()
{
	pbdlen = 0;
	pbd = 0;
}

long long InStream::SetPos(long long pos)
{
	if (pbd && !holdpbd) CompletePBD();

	long long	oldpos = currentfilepos;

	currentfilepos = pos;

	return oldpos;
}

long long InStream::ActPos()
{
	return currentfilepos;
}

long long InStream::Size()
{
	return size;
}

#endif